## The SVM code using CSS functions

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split as tts
from sklearn.metrics import accuracy_score
from sklearn.utils import shuffle
import psycopg2
import datetime

C = 0.01
learning_rate = 0.9

# https://towardsdatascience.com/svm-implementation-from-scratch-python-2db2fc52e5c2#d7d8

def cost_function(W, X, y):
    N = X.shape[0]
    d = 1- y * (X@W)
    d[d < 0] = 0
    hinge_loss = C * (np.sum(d)/N)
    cost = 1/2 * (W@W) + hinge_loss
    return cost


def gradient_function(W, X, y):
    if type(y) == np.float64:
        y = np.array([y])
        X = np.array([X])
    d = 1 - y * (X@W)
    dw = np.zeros(len(W))
    for i, j in enumerate(d):
        if max(0, j) == 0:
            di = W
        else:
            di = W-C*y[i] * X[i]
        dw += di
    dw = dw/len(y)
    return dw


def sgd(features, outputs):
    max_runs = 200
    weights = np.zeros(features.shape[1])
    nth = 0
    prev_cost = float("inf")
    cost_threshold = 0.01
    for runs in range(1, max_runs):
        X, Y = shuffle(features, outputs)
        for i, x in enumerate(X):
            ascent = gradient_function(weights, x, Y[i])
            weights = weights - (learning_rate * ascent)
        # convergence check on 2^nth epoch
        if runs == 2 ** nth or runs == max_runs - 1:
            cost = cost_function(weights, features, outputs)
            if abs(prev_cost - cost) < cost_threshold * prev_cost:
                return weights
            prev_cost = cost
            nth += 1
    return weights

def init():
    start_time = datetime.datetime.now()

    # read data in pandas (pd) data frame
    conn = psycopg2.connect(
        database="mqp", user='postgres', password='password', host='127.0.0.1', port='5432'
    )
    cursor = conn.cursor()
    cursor.execute("select * from transfusion")
    tuples = cursor.fetchall()
    cursor.close()
    column_names = ['X1', 'X2', 'X3', 'X4', 'Y']
    data = pd.DataFrame(tuples, columns=column_names)
    conn.close()

    map_y = {1: 1.0, 0: -1.0}
    data['Y'] = data['Y'].map(map_y)

    X = data.iloc[:, 0:4]
    Y = data.loc[:, 'Y']

    # insert 1 in every row for intercept b
    X.insert(loc=len(X.columns), column='intercept', value=1)

    # split data into train and test set
    X_train, X_test, y_train, y_test = tts(X, Y, train_size=0.8, random_state=1, shuffle = False)
    W = sgd(X_train.to_numpy(), y_train.to_numpy())
    # testing the model

    y_train_pred = np.array([])
    for i in range(X_train.shape[0]):
        yp = np.sign(np.dot(X_train.to_numpy()[i], W))
        y_train_pred = np.append(y_train_pred, yp)

    y_test_pred = np.array([])
    for i in range(X_test.shape[0]):
        yp = np.sign(np.dot(X_test.to_numpy()[i], W))
        y_test_pred = np.append(y_test_pred, yp)

    print("Accuracy rate: {}".format(accuracy_score(y_test, y_test_pred)))

    end_time = datetime.datetime.now()
    time_diff = (end_time - start_time)
    execution_time = time_diff.total_seconds() * 1000
    print("Time: ", execution_time, "ms")



init()



